/******************************************************************************************

   Triplets.java - Triplets board game
   Version 2.0
   www.mazeworks.com
   
   Copyright (c) 2002 Robert Kirkland
   All Rights Reserved.

   Permission is hereby granted, free of charge, to any person obtaining a copy 
   of this software and associated documentation files (the "Software"), to
   deal in the Software without restriction, including without limitation the
   rights to use, copy, modify, merge, publish, distribute, and/or sell copies
   of the Software, and to permit persons to whom the Software is furnished to
   do so, provided that the above copyright notice and this permission notice
   appear in all copies of the Software and that both the above copyright
   notice and this permission notice appear in supporting documentation.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER INCLUDED IN THIS NOTICE BE LIABLE FOR
   ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Except as contained in this notice, the name of the copyright holder shall
   not be used in advertising or otherwise to promote the sale, use or other
   dealings in this Software without prior written authorization of the
   copyright holder.

******************************************************************************************/

import java.applet.* ;
import java.awt.* ;
import java.awt.image.* ;
import java.net.* ;
import java.util.* ;
	
/******** TRIPLETS ********
	main class: initializes applet panel, 
	starts games, handles control panel & moves
*/	
public class Triplets extends Applet implements Runnable {
   // applet info
   public String getAppletInfo() {
      return "Triplets 2.0 - Copyright (c) 2002 Robert Kirkland"; }
   String parameterInfo[][] = {
      {"bgColor",      "integer", "Background color (24-bit RGB hex int)"},
      {"title",        "string",  "Applet title text"},
      {"newGame",      "string",  "New Game button text"},
      {"level",        "string",  "Level label text"},
      {"beginner",     "string",  "Beginner level text"},
      {"intermediate", "string",  "Intermediate level text"},
      {"expert",       "string",  "Expert level text"},
      {"board",        "string",  "Board label text"},
      {"bdWidth",      "string",  "Board Width label text"},
      {"bdHeight",     "string",  "Board Height label text"},
      {"board3",       "string",  "Board size 3 text"},
      {"board4",       "string",  "Board size 4 text"},
      {"board5",       "string",  "Board size 5 text"},
      {"board6",       "string",  "Board size 6 text"},
      {"board7",       "string",  "Board size 7 text"},
      {"board8",       "string",  "Board size 8 text"},
      {"board9",       "string",  "Board size 9 text"},
      {"board10",      "string",  "Board size 10 text"},
      {"board11",      "string",  "Board size 11 text"},
      {"board12",      "string",  "Board size 12 text"},
      {"playerFirst",  "string",  "Player First checkbox text"},
      {"yourMove",     "string",  "Player Move status text"},
      {"thinking",     "string",  "Computer Thinking status text"},
      {"playerWin",    "string",  "Player Win status text"},
      {"tripletWin",   "string",  "Computer Win status text"} } ;
   public String[][] getParameterInfo() { return parameterInfo ; }

   static final Color DF_BG_COLOR=Color.white ;  // default background color
   static final String DF_TITLE   ="TRIPLETS",   // default text strings
                       DF_NEW_GAME="NEW GAME",
                       DF_LEVEL   ="LEVEL",
                       DF_BOARD   ="BOARD ",
                       DF_WIDTH   ="Width ",
                       DF_HEIGHT  ="Height ",
                       DF_BD3 =" 3",
                       DF_BD4 =" 4",
                       DF_BD5 =" 5",
                       DF_BD6 =" 6",
                       DF_BD7 =" 7",
                       DF_BD8 =" 8",
                       DF_BD9 =" 9",
                       DF_BD10="10",
                       DF_BD11="11",
                       DF_BD12="12",
                       DF_BEGINNER    ="Beginner",  
                       DF_INTERMED    ="Intermediate",
                       DF_EXPERT      ="Expert",
                       DF_PLAYER_FIRST="Player first",
                       DF_YOUR_MOVE   =" Your move..",
                       DF_THINKING    =" Thinking..",
                       DF_PLAYER_WIN  =" You win!",
                       DF_TRIPLET_WIN =" Game Over!" ;
   static final int bcWidth=360, bcHeight=360 ;
   protected String TITLE, NEW_GAME, LEVEL, BOARD, WIDTH, HEIGHT,
                    BD3, BD4, BD5, BD6, BD7, BD8, BD9, BD10, BD11, BD12,
                    BEGINNER, INTERMED, EXPERT, PLAYER_FIRST, YOUR_MOVE, THINKING,
                    PLAYER_WIN, TRIPLET_WIN ;
   protected Color BG_COLOR ;
   private Point winSquares[] ;
   private int ply ;
   private boolean playerTurn, gameOver ;
   private Image gifletsImage ;
   private Board bd ;
   private MoveGen mg ;
   private AI ulam ;
   private BoardCanvas bc ;
   private ControlPanel cp ;
   private Thread ulamMoveThread ;
	
   public void init() {
      // --- get applet parameters ---
      // background color
      String bgColorString = getParameter("bgColor") ;
      try { BG_COLOR = bgColorString==null ? DF_BG_COLOR : new Color(Integer.parseInt(bgColorString,16)) ; } 
      catch (NumberFormatException e) {
         BG_COLOR = DF_BG_COLOR ;
         System.out.println("ERROR: Invalid bgColor parameter.") ;
      }
      // title
      TITLE = getParameter("title") ; if (TITLE==null) TITLE = new String(DF_TITLE) ;
      // New Game button
      NEW_GAME = getParameter("newGame") ; if (NEW_GAME==null) NEW_GAME = new String(DF_NEW_GAME) ;
      // Level text
      LEVEL = getParameter("level") ; if (LEVEL==null) LEVEL = new String(DF_LEVEL) ;
      BEGINNER = getParameter("beginner") ; if (BEGINNER==null) BEGINNER = new String(DF_BEGINNER) ;
      INTERMED = getParameter("intermediate") ; if (INTERMED==null) INTERMED = new String(DF_INTERMED) ;
      EXPERT = getParameter("expert") ; if (EXPERT==null) EXPERT = new String(DF_EXPERT) ;
      // Board size text
      BOARD = getParameter("board") ; if (BOARD==null) BOARD = new String(DF_BOARD) ;
      WIDTH = getParameter("bdWidth") ; if (WIDTH==null) WIDTH = new String(DF_WIDTH) ;
      HEIGHT = getParameter("bdHeight") ; if (HEIGHT==null) HEIGHT = new String(DF_HEIGHT) ;
      BD3 = getParameter("board3") ; if (BD3==null) BD3 = new String(DF_BD3) ;
      BD4 = getParameter("board4") ; if (BD4==null) BD4 = new String(DF_BD4) ;
      BD5 = getParameter("board5") ; if (BD5==null) BD5 = new String(DF_BD5) ;
      BD6 = getParameter("board6") ; if (BD6==null) BD6 = new String(DF_BD6) ;
      BD7 = getParameter("board7") ; if (BD7==null) BD7 = new String(DF_BD7) ;
      BD8 = getParameter("board8") ; if (BD8==null) BD8 = new String(DF_BD8) ;
      BD9 = getParameter("board9") ; if (BD9==null) BD9 = new String(DF_BD9) ;
      BD10 = getParameter("board10") ; if (BD10==null) BD10 = new String(DF_BD10) ;
      BD11 = getParameter("board11") ; if (BD11==null) BD11 = new String(DF_BD11) ;
      BD12 = getParameter("board12") ; if (BD12==null) BD12 = new String(DF_BD12) ;
      // Player First checkbox
      PLAYER_FIRST = getParameter("playerFirst") ; if (PLAYER_FIRST==null) PLAYER_FIRST = new String(DF_PLAYER_FIRST) ;
      // status text
      YOUR_MOVE = getParameter("yourMove") ; if (YOUR_MOVE==null) YOUR_MOVE = new String(DF_YOUR_MOVE) ;
      THINKING = getParameter("thinking") ; if (THINKING==null) THINKING = new String(DF_THINKING) ;
      PLAYER_WIN = getParameter("playerWin") ; if (PLAYER_WIN==null) PLAYER_WIN = new String(DF_PLAYER_WIN) ;
      TRIPLET_WIN = getParameter("tripletWin") ; if (TRIPLET_WIN==null) TRIPLET_WIN = new String(DF_TRIPLET_WIN) ;

      // load images
      MediaTracker tracker = new MediaTracker(this) ;
      URL url = getCodeBase() ;
      gifletsImage = getImage(url,"giflets.gif") ;
      tracker.addImage(gifletsImage,0) ;
      try { tracker.waitForAll(); }
      catch (InterruptedException e) {}
      // create board canvas and control panel
      cp = new ControlPanel(this) ;
      bc = new BoardCanvas(bcWidth,bcHeight,gifletsImage,cp,this) ;
      bc.resize(bcWidth,bcHeight) ;
      Panel p1 = new Panel() ;
      p1.setBackground(Color.lightGray) ;
      p1.setLayout(new FlowLayout(FlowLayout.CENTER,5,5)) ;
      Panel p2 = new Panel() ;
      p2.setBackground(Color.gray) ;
      p2.setLayout(new FlowLayout(FlowLayout.CENTER,2,2)) ;
      p1.add(cp) ;
      p2.add(p1) ;
      // initialize applet layout
      setBackground(BG_COLOR) ;
      setLayout(new FlowLayout(FlowLayout.CENTER,10,10)) ;
      repaint() ;
      add(bc) ;
      add(p2) ;
      validate() ;
      // force class file load
      Position p = new Position() ;
      Move m = new Move() ;
		newGame() ;
	}
	void newGame() {
      stop() ;
		System.gc() ; 
      gameOver = false ; 
      ply = 1 ;
      bd = new Board(bc,cp,this) ;
      mg = new MoveGen(bd) ;
      ulam = new AI(bd,mg,cp,this) ;
      if (cp.isPlayerFirst()) {
         playerTurn = true ;
         cp.setStatus(YOUR_MOVE) ;
      }
      else playerTurn = false ;
      bc.drawBoard(bd) ;
      if (!playerTurn) ulamMoveStart() ;
      bc.requestFocus() ;
   }      
   // player move
   void handleMouseDown(int x,int y) {
      if (!gameOver&&playerTurn) {
         if (bd.makePlayerMove(bc.pixelToSquare(bd,x,y))) {
            // check for win
            if ((winSquares=bd.isWinSquares())!=null) {
               bc.drawBoard(bd,winSquares) ;
               cp.setStatus(PLAYER_WIN) ;
               gameOver = true ;
            }
            else {
               ply++ ;
               playerTurn = false ;
               bc.drawBoard(bd) ;
               cp.setStatus(THINKING) ;
               ulamMoveStart() ;
            }
         }
      }
   }
   // start AI move thread
   public void ulamMoveStart() {
      if (ulamMoveThread == null) {
         ulamMoveThread = new Thread(this) ;
         ulamMoveThread.start() ;
      }
   }
   // kill thread
   public void stop() {
      if (ulamMoveThread != null) {
         ulamMoveThread.stop() ;
         ulamMoveThread = null ;
      }
   }
   // AI move
   public void run() { 
      cp.setStatus(THINKING) ;
      ulam.makeMove(ply) ;
      // check for win
      if ((winSquares=bd.isWinSquares())!=null) {
         bc.drawBoard(bd,winSquares) ;
         cp.setStatus(TRIPLET_WIN) ;
         gameOver = true ;
      }
      else {
         ply++ ;
         playerTurn = true ;
         cp.setStatus(YOUR_MOVE) ;
         bc.drawBoard(bd) ;
      }
      ulamMoveThread = null ;
   }
   void setLevel() { bc.requestFocus() ; }
   int getPly() { return ply ; }
   void sleep(int ms) {
      try { Thread.currentThread().sleep(ms) ; } 
      catch (InterruptedException e) {}
   }
}      
/******** BOARD ********
	initializes board, makes moves, checks for win
*/	
final class Board {
   // boards are arrays of 16-bit integers, one integer 
   // for each row, columns starting from rightmost bit
   static final int MAX_SQUARES=144 ;
   static final short EMPTY=0, COLUMN[]= { 0x800, 0x400, 0x200, 0x100,
                                           0x080, 0x040, 0x020, 0x010,
                                           0x008, 0x004, 0x002, 0x001 } ;
   private int bdCols, bdRows, squareValue[][] ;
   private short bd[] ;
   private BoardCanvas bc ;
   private ControlPanel cp ;
   private Triplets main ;
	
   // constructor
   Board(BoardCanvas bc,ControlPanel cp,Triplets main) {
      this.bc = bc ;
      this.cp = cp ;
      this.main = main ;
      // initialize board
      bdCols = cp.getBoardCols() ;
      bdRows = cp.getBoardRows() ;
      bd = new short[bdRows] ;
      // set square values 
      // higher values towards center
      squareValue = new int[bdCols][bdRows] ;
      int rowVal, colVal ;      
      for (int row=0; row<bdRows; row++) { 
         bd[row] = 0 ; 
         for (int col=0; col<bdCols; col++) {
            if (row<(bdRows/2)) rowVal = row ;
            else rowVal = bdRows-row-1 ;
            if (col<(bdCols/2)) colVal = col ;
            else colVal = bdCols-col-1 ;
            squareValue[col][row] = colVal + rowVal ;
         }
      }
   }
   int getBoardRows() { return bdRows ; }
   int getBoardCols() { return bdCols ; }
   int getRowValue(int row) { return bd[row] ; }
   int getSquareValue(int col,int row) { return squareValue[col][row] ; }
   boolean isPiece(Point p) { return isPiece(p.x,p.y) ; }
   boolean isPiece(int col,int row) { return (bd[row]&COLUMN[col])!=EMPTY ; }
   void setPiece(Point p) { setPiece(p.x,p.y) ; }
   void setPiece(int col,int row) { bd[row] |= COLUMN[col] ; }
   void setEmpty(Point p) { setEmpty(p.x,p.y) ; }
   void setEmpty(int col,int row) { bd[row] &= ~COLUMN[col] ; }

   private boolean isValidSquare(Point p) {
      if (p!=null) return isValidSquare(p.x,p.y) ;
      return false ;
   }
	private boolean isValidSquare(int col,int row) { 
      if ( (col>=0)&&(col<bdCols) &&
           (row>=0)&&(row<bdRows) &&
           (!isPiece(col,row)) ) return true ;
      return false ;
   }
   boolean makePlayerMove(Point p) {
      if (isValidSquare(p)) {
         setPiece(p) ;
         return true;
      }
      return false ;
   }
   void makeAIMove(Point p) {
      setPiece(p) ;
      bc.drawBoard(this) ;
      main.sleep(400) ;
      setEmpty(p) ;
      bc.drawBoard(this) ;
      main.sleep(400) ;
      setPiece(p) ;
   }
   // check board for win, returning winning 3 squares if found
   Point[] isWinSquares() {
      for (int row=0; row<bdRows; row++) {
         for (int col=0; col<bdCols; col++) {
            if (isPiece(col,row)) {
               // East
               if ((col<bdCols-2)&&isPiece(col+1,row)&&isPiece(col+2,row)) {
                  Point winSquares[] = { new Point(col,row),
                                         new Point(col+1,row),
                                         new Point(col+2,row) } ;
                  return winSquares ;
               }
               // SouthEast
               if ((col<bdCols-2)&&(row<bdRows-2)&&
                    isPiece(col+1,row+1)&&isPiece(col+2,row+2)) {
                  Point winSquares[] = { new Point(col,row),
                                         new Point(col+1,row+1),
                                         new Point(col+2,row+2) } ;
                  return winSquares ;
               }
               // South
               if ((row<bdRows-2)&&isPiece(col,row+1)&&isPiece(col,row+2)) {
                  Point winSquares[] = { new Point(col,row),
                                         new Point(col,row+1),
                                         new Point(col,row+2) } ;
                  return winSquares ;
               }
               // SouthWest
               if ((col>1)&&(row<bdRows-2)&&
                    isPiece(col-1,row+1)&&isPiece(col-2,row+2)) {
                  Point winSquares[] = { new Point(col,row),
                                         new Point(col-1,row+1),
                                         new Point(col-2,row+2) } ;
                  return winSquares ;
               }
            }
         }
      }
      return null ;
   }
   // check entire board for win
   boolean isWin() {
      for (int row=0; row<bdRows; row++) {
         for (int col=0; col<bdCols; col++) {
            if (isPiece(col,row)) {
               // East
               if ((col<bdCols-2)&&isPiece(col+1,row)&&isPiece(col+2,row))
                  return true ;
               // SouthEast
               if ((col<bdCols-2)&&(row<bdRows-2)&&
                    isPiece(col+1,row+1)&&isPiece(col+2,row+2))
                  return true ;
               // South
               if ((row<bdRows-2)&&isPiece(col,row+1)&&isPiece(col,row+2))
                  return true ;
               // SouthWest
               if ((col>1)&&(row<bdRows-2)&&
                    isPiece(col-1,row+1)&&isPiece(col-2,row+2))
                  return true ;
            }
         }
      }
      return false ;
   }
   // check for winning move
   boolean isWin(int col,int row) {
      // East-West
      if ( ((col>1)&&isPiece(col-1,row)&&isPiece(col-2,row)) ||
           ((col>0)&&(col<bdCols-1)&&isPiece(col-1,row)&&isPiece(col+1,row)) ||
           ((col<bdCols-2)&&isPiece(col+1,row)&&isPiece(col+2,row)) )
         return true ;
      // North-South
      if ( ((row<bdRows-2)&&isPiece(col,row+1)&&isPiece(col,row+2)) ||
           ((row>0)&&(row<bdRows-1)&&isPiece(col,row-1)&&isPiece(col,row+1)) ||
           ((row>1)&&isPiece(col,row-1)&&isPiece(col,row-2)) )
         return true ;
      // NorthWest-SouthEast
      if ( ((col<bdCols-2)&&(row<bdRows-2)&&isPiece(col+1,row+1)&&isPiece(col+2,row+2)) ||
           ((col>0)&&(col<bdCols-1)&&(row>0)&&(row<bdRows-1)&&
             isPiece(col-1,row-1)&&isPiece(col+1,row+1)) ||
           ((col>1)&&(row>1)&&isPiece(col-1,row-1)&&isPiece(col-2,row-2)) )
         return true ;
      // NorthEast-SouthWest
      if ( ((col>1)&&(row<bdRows-2)&&isPiece(col-1,row+1)&&isPiece(col-2,row+2)) ||
           ((col>0)&&(col<bdCols-1)&&(row>0)&&(row<bdRows-1)&&
             isPiece(col-1,row+1)&&isPiece(col+1,row-1)) ||
           ((col<bdCols-2)&&(row>1)&&isPiece(col+1,row-1)&&isPiece(col+2,row-2)) )
         return true ;
      
      return false ;
   }
   // check for "safe" move (does not create a couplet)
   boolean isSafe(int col,int row) {
      // East-West
      if ( ((col>1)&&(isPiece(col-1,row)||isPiece(col-2,row))) ||
           ((col>0)&&(col<bdCols-1)&&(isPiece(col-1,row)||isPiece(col+1,row))) ||
           ((col<bdCols-2)&&(isPiece(col+1,row)||isPiece(col+2,row))) )
         return false ;
      // North-South
      if ( ((row<bdRows-2)&&(isPiece(col,row+1)||isPiece(col,row+2))) ||
           ((row>0)&&(row<bdRows-1)&&(isPiece(col,row-1)||isPiece(col,row+1))) ||
           ((row>1)&&(isPiece(col,row-1)||isPiece(col,row-2))) )
         return false ;
      // NorthWest-SouthEast
      if ( ((col<bdCols-2)&&(row<bdRows-2)&&(isPiece(col+1,row+1)||isPiece(col+2,row+2))) ||
           ((col>0)&&(col<bdCols-1)&&(row>0)&&(row<bdRows-1)&&
             (isPiece(col-1,row-1)||isPiece(col+1,row+1))) ||
           ((col>1)&&(row>1)&&(isPiece(col-1,row-1)||isPiece(col-2,row-2))) )
         return false ;
      // NorthEast-SouthWest
      if ( ((col>1)&&(row<bdRows-2)&&(isPiece(col-1,row+1)||isPiece(col-2,row+2))) ||
           ((col>0)&&(col<bdCols-1)&&(row>0)&&(row<bdRows-1)&&
             (isPiece(col-1,row+1)||isPiece(col+1,row-1))) ||
           ((col<bdCols-2)&&(row>1)&&(isPiece(col+1,row-1)||isPiece(col+2,row-2))) )
         return false ;
      
      return true ;
   }
}         
/******** AI ********
   generates computer moves
*/   
final class AI {
   static final int PLAYER_WIN=0, UNKNOWN=1, COMPUTER_WIN=2,
                    PLAYER=0, COMPUTER=1 ;
   private Hashtable transTable = new Hashtable(1009) ;
   private int cutoff, bdSize ;
   private byte[][] moveList ;
   private Point mv ;
   private Board bd ;
   private MoveGen mg ;
   private ControlPanel cp ;
   private Triplets main ;

   // constructor
   AI(Board bd,MoveGen mg,ControlPanel cp,Triplets main) {
      this.bd = bd ;
      this.mg = mg ;
      this.cp = cp ;
      this.main = main ;
   }
   // make move based on ply and game level
   boolean makeMove(int ply) { 
      mv = null ;
      switch (cp.getLevelIndex()) {
         case cp.BEG: mv = bestMove(false) ; 
                      break ;
         case cp.INT: if (ply==1) mv = openingMove() ; 
                      else mv = bestMove(true) ; 
                      break ;
         case cp.EXP: 
            if (ply==1) mv = openingMove() ; 
            else {
               // set ply cutoff to kick in alpha-beta search
               bdSize = bd.getBoardRows() * bd.getBoardCols() ;
               if      (bdSize< 64)  cutoff = 1 ;
               else if (bdSize==64)  cutoff = 2 ;
               else if (bdSize< 81)  cutoff = 3 ;
               else if (bdSize==81)  cutoff = 4 ;
               else if (bdSize< 90)  cutoff = 5 ;
               else if (bdSize< 100) cutoff = 6 ;
               else if (bdSize==100) cutoff = 7 ;
               else if (bdSize<=110) cutoff = 8 ;
               else if (bdSize<=121) cutoff = 9 ;
               else if (bdSize==132) cutoff = 11 ;
               else cutoff = 14 ;
               if (ply>cutoff)
                  mv = abSearch(COMPUTER,PLAYER_WIN,COMPUTER_WIN,0).getSquare() ;
               if (mv==null) { 
                  mv = bestMove(true) ;
                  // pretend we're thinking about it
                  main.sleep(randomInt(2000)) ;
               }
            }
            break ;
      }
      bd.makeAIMove(mv) ; 
      return true ;
   }
   private Point bestMove(boolean bestOnly) {
      // if bestOnly, return the highest-value move,
      // otherwise choose one at random
      moveList = mg.genMoveList() ;
      int rnd = bestOnly? 0 : randomInt(moveList.length) ;
      return new Point(moveList[rnd][mg.COL],moveList[rnd][mg.ROW]) ;
   }
   // open on a random central square
   private Point openingMove() {
      int col=0, row=0 ;

      switch (randomInt(4)) {
         case 0: col=(bd.getBoardCols()-1)/2 ;
                 row=(bd.getBoardRows()-1)/2 ;
                 break ;
         case 1: col=(bd.getBoardCols()-1)/2 ;
                 row=bd.getBoardRows()/2 ;
                 break ;
         case 2: col=bd.getBoardCols()/2 ;
                 row=(bd.getBoardRows()-1)/2 ;
                 break ;
         case 3: col=bd.getBoardCols()/2 ;
                 row=bd.getBoardRows()/2 ;
                 break ; 
      }
      return new Point(col,row) ;
   }
   // alpha-beta -- search recursively to end of game, 
   // returning forced-win move if found, null otherwise
   private Move abSearch(int side,int alpha,int beta,int depth) {
      int col, row, otherSide, baseCase, value ;
      Point transData=null, savedMove=null ;
      Position thisPosition = new Position(bd) ;
      Move reply ;
      
      // base to end recursion
      baseCase = staticEval(side) ;
      if (baseCase!=UNKNOWN) return new Move(baseCase) ;
      // check if position in transposition table
      if (depth==0) transTable.clear() ;
      else if ((transData=(Point)transTable.get(thisPosition))!=null) {
         if (transData.x==side) return new Move(transData.y) ;
         else {
            // swap values if data is for other side
            if (transData.y==COMPUTER_WIN) return new Move(PLAYER_WIN) ;
            else return new Move(COMPUTER_WIN) ;
         }
      }
      // set initial alpha-beta values
      if (side==COMPUTER) { otherSide = PLAYER ; value = alpha ; }
      else { otherSide = COMPUTER ; value = beta ; }
      // loop for all generated moves
      byte ml[][] = mg.genMoveList(depth) ;
      for (int index=0; index<ml.length; index++) {
         col = ml[index][mg.COL] ;
         row = ml[index][mg.ROW] ;
         bd.setPiece(col,row) ;
         // recursive search
         reply = abSearch(otherSide,alpha,beta,depth+1) ;
         bd.setEmpty(col,row) ;
         
         if ( ((side==COMPUTER)&&(reply.getValue()>value))||
              ((side==PLAYER)&&(reply.getValue()<value)) ) {
            // save better move 
            if (side==COMPUTER) alpha = value = reply.getValue() ;
            else beta = value = reply.getValue() ;
            savedMove = new Point(col,row) ;
            // refutation
            if (alpha >= beta) break ;
         }
      }
      // save position-side-value in transposition table
      transTable.put(thisPosition,new Point(side,value)) ;
      return new Move(savedMove,value) ;
   }
   // since we're searching to end of game,
   // the position evaluator only checks for win
   private int staticEval(int side) {
      if (bd.isWin()) return (side!=COMPUTER)?COMPUTER_WIN:PLAYER_WIN ;
      else return UNKNOWN ;
   }
   // returns random integer from 0 to n-1 inclusive
   int randomInt(int n) { return (int)(n*Math.random()) ; }

}
/******** MOVEGEN ********
   move generator
*/   
final class MoveGen {
   static final int WIN_VALUE=40, SAFE_VALUE=20, OPEN_SQUARE=0,
                    CRV=3, COL=0, ROW=1, VALUE=2, MAX_MOVES=8 ;
   private int temp[]=new int[CRV], intList[][], totalCount, 
               winCount, safeCount ;
   private Board bd ;

   // constructor
   MoveGen(Board bd) {
      this.bd = bd ;
      intList = new int[bd.MAX_SQUARES][CRV] ;
   }
   // create movelist for alpha-beta and random move
   byte[][] genMoveList() { return genMoveList(-1) ; }
   byte[][] genMoveList(int depth) {
      // first generate int list of all moves with values
      totalCount = winCount = safeCount = 0 ;
      for (int col=0; col<bd.getBoardCols(); col++) {
         for (int row=0; row<bd.getBoardRows(); row++) {
            if (!bd.isPiece(col,row)) {
               intList[totalCount][COL] = col ;
               intList[totalCount][ROW] = row ;
               if (bd.isWin(col,row)) {
                  intList[totalCount][VALUE] = WIN_VALUE ;
                  winCount++ ;
               }
               else {
                  if (bd.isSafe(col,row)) {
                     intList[totalCount][VALUE] = SAFE_VALUE ;
                     safeCount++ ;
                  }
                  else intList[totalCount][VALUE] = OPEN_SQUARE ;
               }
               // add value of square
               intList[totalCount][VALUE] += bd.getSquareValue(col,row) ;
               totalCount++ ;
            }
         }
      }
      // now shellsort integer list by value
      for (int h=intList.length/2; h>0; h/=2) {
         for (int i=h; i<intList.length; i++) {
            for (int k=0; k<CRV; k++) temp[k] = intList[i][k] ; 
            int j = i ;
            for ( ; j>=h && temp[VALUE]>intList[j-h][VALUE] ; j-=h)
               for (int k=0; k<CRV; k++) intList[j][k] = intList[j-h][k] ;
            for (int k=0; k<CRV; k++) intList[j][k] = temp[k] ;
         }
      }
      // if not alpha-beta, save all moves of the best type, 
      // otherwise only max moves with the highest value
      if (depth<0) {
         if (winCount!=0) totalCount = winCount ;
         else if (safeCount!=0) totalCount = safeCount ;
      }
      else totalCount = Math.min(totalCount,MAX_MOVES) ;
      // convert to byte list and return
      byte byteList[][] = new byte[totalCount][CRV-1] ;
      for (int i=0; i<totalCount; i++) {
         byteList[i][COL] = (byte)intList[i][COL] ;
         byteList[i][ROW] = (byte)intList[i][ROW] ;
      }
      return byteList ;
   }
}
/******** MOVE ********
   data for a single move
*/   
final class Move {
   private int value ;
   private Point square ;
   
   // constructors
   Move() { }
   Move(int value) { 
      this.square = null ;
      this.value = value ;
   }
   Move(Point square,int value) {
      if (square!=null) this.square = new Point(square.x,square.y) ;
      this.value = value ;
   }
   public boolean equals(Object o) {
      Move mv = (Move)o ;
      if (square.equals(mv.square)) return true ;
      return false ;
   }
   Point getSquare() { return square ; }
   int getValue() { return value ; }
}
/******** POSITION ********
   boards positions used as key in transposition table 
*/   
final class Position {
   private short[] pos ;
   private int rows ;
   
   // constructor
   Position() { }
   Position(Board bd) {
      rows=bd.getBoardRows() ;
      pos = new short[rows] ;
      for (int i=0; i<rows; i++)
         pos[i] = (short)bd.getRowValue(i) ;
   }
   // hash the position
   public int hashCode() {
      int hashValue=0 ;
      for (int i=0; i<rows; i++) {
         hashValue += pos[i]*(i+1) ;
      }
      return hashValue ;
   }
   public boolean equals(Object otherPos) {
      for (int i=0; i<rows; i++)
         if (pos[i] != ((Position)otherPos).pos[i]) return false ;
      return true ;
   }
}
/******** BOARD CANVAS ********
	controls board graphics, handles board mouse events 
*/	
final class BoardCanvas extends Canvas {
   final static int squareCoord[]={0,0,28,28}, winSquareCoord[]={28,0,28,28},
                    blackPieceCoord[]={56,0,25,25}, SQUARE=28, OFFSET=5 ;
   final static Color frameColor=new Color(204,153,0),
                      frameHighColor=new Color(255,204,102) ;
   private int canvasWidth, canvasHeight ;
	private Image bufferImage, gifletsImage, squareImage, winSquareImage, 
                 pieceImage ;
   private MediaTracker tracker ;
	private Graphics buffer ;
   private ControlPanel cp ;
	private Triplets main ;

	BoardCanvas(int canvasWidth,int canvasHeight,Image gifletsImage,
               ControlPanel cp,Triplets main) { 
      this.canvasWidth = canvasWidth ;
      this.canvasHeight = canvasHeight ;
      this.gifletsImage = gifletsImage ;
      this.cp = cp ;
      this.main = main ;
      // extract images from giflets GIF
      tracker = new MediaTracker(main) ;
      squareImage = extractImage(squareCoord) ;
      winSquareImage = extractImage(winSquareCoord) ;
      pieceImage = extractImage(blackPieceCoord) ;
      try { tracker.waitForAll() ; }
      catch (InterruptedException e) {}
   }
   // draw board and pieces
	void drawBoard(Board bd) { drawBoard(bd,null) ; }
   void drawBoard(Board bd,Point winSquares[]) {
      int x, y, bdCols=bd.getBoardCols(), bdRows=bd.getBoardRows(), 
          boardXOffset=getXOffset(bd), boardYOffset=getYOffset(bd) ;

      if (buffer==null) {
         bufferImage = createImage(canvasWidth,canvasHeight) ;
         buffer = bufferImage.getGraphics() ;
      }
      buffer.setColor(main.BG_COLOR) ;
      buffer.fillRect(0,0,canvasWidth,canvasHeight) ;
      // draw frame
      x = (bdCols * SQUARE) + 1 ;
      y = (bdRows * SQUARE) + 1 ;
      buffer.setColor(frameColor) ;
      buffer.drawRect(boardXOffset-3,boardYOffset-3,x+4,y+4) ;
      buffer.drawRect(boardXOffset-1,boardYOffset-1,x,y) ;
      buffer.setColor(frameHighColor) ;
      buffer.drawRect(boardXOffset-2,boardYOffset-2,x+2,y+2) ;
      // draw the board and pieces
      for (int col=0; col<bdCols; col++) {
         for (int row=0; row<bdRows; row++) {
            x = boardXOffset + (col*SQUARE) ;
            y = boardYOffset + (row*SQUARE) ;
            if (winSquares==null) buffer.drawImage(squareImage,x,y,this) ;
            else {
               for (int i=0; i<3; i++) {
                  if ((col==winSquares[i].x)&&(row==winSquares[i].y)) {
                     buffer.drawImage(winSquareImage,x,y,this) ;
                     break ;
                  }
                  else buffer.drawImage(squareImage,x,y,this) ;
               }
            }
            if (bd.isPiece(col,row))
               buffer.drawImage(pieceImage,x+2,y+2,this) ;
         }
      }
      repaint() ;
   }
	public void paint(Graphics g) { update(g) ; }
	public void update(Graphics g) { g.drawImage(bufferImage,0,0,this) ; }
	
	// board canvas mouse event handler
   public boolean mouseDown(Event e,int x,int y) {   
      main.handleMouseDown(x,y) ;
      return true ; 
   }
   Point pixelToSquare(Board bd,int x,int y) {
      int boardXOffset=getXOffset(bd), boardYOffset=getYOffset(bd) ;

      if ((x>boardXOffset)&&(y>boardYOffset)) {
         return new Point((x-boardXOffset)/SQUARE,(y-boardYOffset)/SQUARE) ;
      }
      return null ;
   }
   private int getXOffset(Board bd) { return (canvasWidth-(bd.getBoardCols()*SQUARE))/2 ; }
   private int getYOffset(Board bd) { return (canvasHeight-(bd.getBoardRows()*SQUARE))/2 ; }
   // extract GIF images
   private Image extractImage(int[] xyCoord) {
      Image newImage ;
      ImageFilter filter;
      ImageProducer producer;

      filter = new CropImageFilter(xyCoord[0],xyCoord[1],xyCoord[2],xyCoord[3]);
      producer = new FilteredImageSource(gifletsImage.getSource(), filter);
      newImage = main.createImage(producer);
      tracker.addImage(newImage,0);
      return newImage ;
   }
}
/******** CONTROL PANEL ********
   the UI components
*/   
final class ControlPanel extends Panel {
   static final int BEG=0, INT=1, EXP=2,
                    boardSizes[]= {3,4,5,6,7,8,9,10,11,12} ;
   static final Font titleFont=new Font("SansSerif", Font.BOLD, 14),
                     textFont=new Font("SansSerif", Font.PLAIN, 12),
                     monoFont=new Font("Courier", Font.PLAIN, 12),
                     statusFont=new Font("Dialog", Font.PLAIN, 12) ;
   private Button bNewGame ;
   private Choice cBoardWidth, cBoardHeight, cLevel ;
   private Checkbox cbPlayerFirst ;
   private TextField tfStatus ;
   private Label title ;
   private Triplets main ;

   // constructor
   ControlPanel(Triplets main) {
      this.main = main ;
      String sizeList[] = { main.BD3, main.BD4, main.BD5, main.BD6, main.BD7, 
                            main.BD8, main.BD9, main.BD10, main.BD11, main.BD12 } ;
      // main layout
      setLayout(new GridLayout(0,1,0,5)) ;
      setBackground(Color.lightGray) ;
      setFont(textFont) ;
      // Title
      title = new Label(main.TITLE,Label.CENTER) ;
      title.setFont(titleFont) ;
      add(title) ;
      add(new Label(" ")) ;
      // New Game button
      add(bNewGame = new Button(main.NEW_GAME)) ;
      // Level pulldown
      add(new Label(main.LEVEL,Label.CENTER)) ;
      add(cLevel = new Choice()) ;
      cLevel.addItem(main.BEGINNER);
      cLevel.addItem(main.INTERMED);
      cLevel.addItem(main.EXPERT);
      cLevel.select(main.BEGINNER) ;
      // Board Width pull-down panel
      add(new Label(main.BOARD,Label.CENTER)) ;
      Panel bwp = new Panel() ;
      bwp.setLayout(new FlowLayout(FlowLayout.RIGHT,3,0)) ;
      bwp.add(new Label(main.WIDTH,Label.CENTER)) ;
      bwp.add(cBoardWidth = new Choice()) ;
      cBoardWidth.setFont(monoFont) ;
      for (int i=0; i<sizeList.length; i++)
         cBoardWidth.addItem(sizeList[i]);
      cBoardWidth.select(main.BD6) ;
      add(bwp) ;
      // Board Height pull-down panel
      Panel bhp = new Panel() ;
      bhp.setLayout(new FlowLayout(FlowLayout.RIGHT,3,0)) ;
      bhp.add(new Label(main.HEIGHT,Label.CENTER)) ;
      bhp.add(cBoardHeight = new Choice()) ;
      cBoardHeight.setFont(monoFont) ;
      for (int i=0; i<sizeList.length; i++)
         cBoardHeight.addItem(sizeList[i]);
      cBoardHeight.select(main.BD6) ;
      add(bhp) ;
      // Player Move First checkbox
      add(cbPlayerFirst = new Checkbox(main.PLAYER_FIRST,null,true)) ;
      // Status field
      tfStatus = new TextField(10) ;
      tfStatus.setFont(statusFont) ;
      tfStatus.setForeground(Color.white) ;
      tfStatus.setBackground(Color.black) ;
      tfStatus.setEditable(false) ;
      add(tfStatus) ;
      validate() ;
   }
   // handle control panel events
   public boolean action(Event e,Object o) { 
      if (main.NEW_GAME.equals(o)) main.newGame() ;
      else if (e.target==cBoardWidth) main.newGame() ;
      else if (e.target==cBoardHeight) main.newGame() ;
      else if (e.target==cLevel) main.setLevel() ;
      else if (e.target==cbPlayerFirst) main.newGame() ;
      return true ;
   }
   void setStatus(String s) { tfStatus.setText(s) ; }
   int getBoardColsIndex() { return cBoardWidth.getSelectedIndex() ; }
   int getBoardRowsIndex() { return cBoardHeight.getSelectedIndex() ; }
   int getBoardCols() { return boardSizes[getBoardColsIndex()] ; }
   int getBoardRows() { return boardSizes[getBoardRowsIndex()] ; }
   int getLevelIndex() { return cLevel.getSelectedIndex() ; }
   boolean isPlayerFirst() { return cbPlayerFirst.getState() ; }
}